#include "shortknot.h"
#include<iostream>
#include"point.h"
#include"vector.h"
#include<iomanip>
#include<ctype.h>
#include"knots2.h"
// #include"tcknots.h"

using namespace std;

shortknot::shortknot():numSides(1), vertex(1)
{
  /* empty */
}

shortknot::shortknot(unsigned int numberOfSides):
  numSides(numberOfSides), vertex(numberOfSides)
{
  /* empty */
}

shortknot::shortknot(unsigned int numberOfSides, double fillvalue):
  numSides(numberOfSides), vertex(numberOfSides)
{
  for(int i=0; i<numSides; i++){
    for(int j=0; j<3; j++){
      vertex[i][j] = fillvalue;
    }
  }
}

// shortknot::shortknot(unsigned int p, unsigned int q, double bigRadius,
// 	     double littleRadius, int numberOfVertices): 
//              numSides(numberOfVertices), vertex(numberOfVertices)
// {
//   double h=p+q+bigRadius+littleRadius+numberOfVertices;
// }

shortknot::shortknot(const shortknot & source):
  numSides(source.vnum()), vertex(source.vnum())
{
  for (int i=0; i<numSides; i++)
    vertex[i] = source[i];
}

shortknot::shortknot(const knots & source):
  numSides(source.vnum()), vertex(source.vnum())
{
  for (int i=0; i<numSides; i++)
    vertex[i] = source[i];
}

// shortknot::shortknot(const tcknots & source):
//   numSides(source.vnum()), vertex(source.vnum())
// {
//   for (int i=0; i<numSides; i++)
//     vertex[i] = source[i];
// } 

shortknot::~shortknot()
{
  vertex.setSize(0);
}

const point & shortknot::operator [] (int index) const
{
  if (index < 0)
    index += numSides;
  return vertex[index%numSides];
}

point & shortknot::operator [](int index)
{
  if (index < 0)
    index += numSides;
  return vertex[index%numSides];
}

unsigned int shortknot::setSize(unsigned int numberOfSides)
{
  numSides = numberOfSides;
  return vertex.setSize(numberOfSides);
}

unsigned int shortknot::vnum() const
{
  return numSides;
}

void shortknot::clear()
{
  for(int i=0; i<numSides; i++){
    for(int j=0; j<3; j++){
      vertex[i][j] = 0;
    }
  }
}


ostream & operator << (ostream & out, const shortknot & source)
{
  int top = source.vnum();

  out << "{" << top << "}\n";

  out.setf(ios::fixed);
  out.precision(10);

  for (int i=0; i<top; i++)
    out << source[i][0] << ",\t" << source[i][1] << ",\t" 
      << source[i][2] << endl;

  return out;
}

istream & operator >> (istream & in, shortknot & source)
{
  char comma;
  int vnum;

  /* first extract the number of vertices */

  while( ! isdigit(in.peek()) )
    in.get(comma);

  in >> vnum;

  while( (! isdigit(in.peek())) && (in.peek() != '-') )
    in.get(comma);



  source.numSides = vnum;
  source.setSize(vnum);

  for (int i=0; i<vnum; i++)
    {
      while( (! isdigit(in.peek())) && (in.peek() != '-') )
	in.get(comma);

      in >> source[i][0];

      while( (! isdigit(in.peek())) && (in.peek() != '-') )
	in.get(comma);

      in >> source[i][1];

      while( (! isdigit(in.peek())) && (in.peek() != '-') )
	in.get(comma);

      in >> source[i][2];
    }

 
  return in;
}

unsigned int operator ==(shortknot a, shortknot b)
{
  if(a.vnum() != b.vnum())
    return 0;

  int top = a.vnum();

  for (int i=0; i<top; i++)
    for (int j=0; j<3; j++)
      {
	if(a[i][j] != b[i][j])
	  return 0;
      }
  return 1;
}

unsigned int operator !=(shortknot a, shortknot b)
{
  return !(a == b);
}

shortknot operator + (const shortknot & knot1, const shortknot & knot2)
{
  if (knot1.vnum() != knot2.vnum() )
    {
      cerr << "Different number of vertices!\n";
      return knot1;
    }

  int top = knot1.vnum();
  shortknot sum(top);

  for(int i=0; i<top; i++)
    {
      sum[i][0] = knot1[i][0] + knot2[i][0];
      sum[i][1] = knot1[i][1] + knot2[i][1];
      sum[i][2] = knot1[i][2] + knot2[i][2];
    }

  return sum;
}

shortknot operator * (double num, const shortknot & source)
{
  int top = source.vnum();
  shortknot sum(top);

  for(int i=0; i<top; i++)
    {
      sum[i][0] = num * source[i][0];
      sum[i][1] = num * source[i][1];
      sum[i][2] = num * source[i][2];
    }

  return sum;
}  


