#include <stdlib.h>
#include <assert.h>
#include <iostream>
#include"vectors.h"

using namespace std;

vectors::vectors()                               // default constructor 0 elts
     // postcondition: vectors of zero items constructed
{
  myLength = 0; myList = 0;
}
    
vectors::vectors(int size)                       // specify size of vectors
     // postcondition: vectors of size items constructed   
{
  myLength = size;
  myList = new double[size];
  assert(myList != 0);
}
    
vectors::vectors(int size, double fillValue)     // specify size and fill value
     // postcondition: vectors of size items, each initialized to fillValue
     //                constructed    
{
  myLength = size;
  myList = new double [size];
  assert(myList != 0);
  for(int k = 0; k < size; k++){
    myList[k] = fillValue;
  }       
}

vectors::vectors(const vectors & vec)             // copy constructor
     // precondition: double supports assignment
     // postcondition: return copy of vec        
{
  // allocate storage
  
  myList = new double [myLength = vec.myLength];
  assert(myList != 0);
  
  // copy elements
  for(int k = 0; k < vec.myLength; k++)
    {
      myList[k] = vec.myList[k];
    }       
}
    
vectors::~vectors ()                            // free new'd storage
     // postcondition: dynamically allocated storage freed
{
  delete [] myList;
  myList = 0;
  myLength = 0;               // leave in "empty" state
}

vectors & vectors::operator = (const vectors & vec) // overload assignment
     // precondition: Item supports assignment     
     // postcondition: self is assigned vec
{
  if (this != &vec)      // don't assign to self!
    {       
      delete [] myList;            // out with old list, in with new
      myList = new double [myLength = vec.myLength]; 
      assert(myList != 0);
      
      // copy vec
      myLength = vec.myLength;
      for(int k=0; k < myLength; k++)
	{
	  myList[k] = vec.myList[k];
	}
    }
  return *this;           
}
    
int vectors::Length() const                    // capacity of vectors
{
  return myLength;
}

int vectors::length() const
{
  return Length();
}

void vectors::Fill(double fillValue)
     // postcondition: all entries == fillvalue
{
  int k;
  for(k=0; k < myLength; k++)
    {
      myList[k] = fillValue;
    }
}

void vectors::SetSize(int newSize)             // change size dynamically
     // precondition: vectors has room for myLength entries     
     // postcondition: vectors has room for newSize entries
     //                the first myLength of which are copies of original
     //                unless newSize < myLength, then truncated copy occurs       
{
  int numToCopy = newSize < myLength ? newSize : myLength;
  
  // allocate new storage
  double * newList = new double[newSize];
  assert(newList != 0);   // be sure storage allocated
  
  int k;
  for(k=0; k < numToCopy;k++)
    {
      newList[k] = myList[k];
    }
  
  delete [] myList;         // de-allocate old storage
  myLength = newSize;
  myList = newList;
}

void vectors::resize(int newSize)
{
  SetSize(newSize);
}
    
double & vectors::operator [] (int index)      
     // safe indexing, returning reference
     // precondition: 0 <= index < myLength
     // postcondition: return index-th item
     // exception: aborts if index is out-of-bounds
{
  if(index < 0)
    {
      while(index < 0)
	index += myLength;
      return myList[index];
    }
  else if(index >= myLength)
    return myList[index % myLength];
  else 
    return myList[index];
  
  /* if ((unsigned) index >= (unsigned)myLength || index < 0)
     {
     cerr << "Illegal vectors index: " << index
     << " (max = " << myLength-1 << ")" << endl;
     assert(index >= 0);
     assert(index < myLength);
     }
     
     return myList[index];  */
  
}

const double & vectors::operator [] (int index) const // const index 
     // safe indexing, returning const reference to avoid modification
     // precondition: 0 <= index < myLength
     // postcondition: return index-th item
     // exception: aborts if index is out-of-bounds
{
  if(index < 0)
    {
      while(index < 0)
	index += myLength;
      return myList[index];
    }
  else if(index >= myLength)
    return myList[index % myLength];
  else 
    return myList[index];
  
  /* if ((unsigned) index >= (unsigned)myLength || index < 0)
     {
     cerr << "Illegal vectors index: " << index
     << " (max = " << myLength << ")" << endl;
     assert(index >= 0);
     assert(index < myLength);
     }
     
     return myList[index]; */
}

